//////////////////////////////////////////
//										//
// KERNIT - Kerning library				//
// { ALPHA }							//
// 										//
// by Geoff Hinchcliffe 				//
// http://gravitron.com.au/kernit		//
// @gravitron							//
// 										//
// 										//
// Free to use, edit, etc.				//
// 										//
// This version requires jQuery			//
// 										//
//////////////////////////////////////////

// globals
var kernitClass = 'kernit', kernitTracking, fontMeasure, kernitTag, loadfont, fonts, fontRegister=0, defaultfont = '100px Times, serif', infoTimer, appTimer, returnFunc, retinaDisplay = 1;

$(document).ready(function(){
	goKernit();
});

// 0. GO!
function goKernit(returnFunction) {
	loadMessage('Loading fonts...');
	if (getStyle('.retinadisplay', 'font-size')!=null) retinaDisplay = 0.1;
	if (returnFunction!==undefined) returnFunc = returnFunction;
	setTimeout( function(){ checkStyles(); } , 10);// firefox needs this small delay
}

// 1. Check that external style sheet containing kernit styles has loaded.
function checkStyles() {
	loadfont = getStyle('.kernit_load_fonts', 'font-family');
	if (loadfont) {
		fonts = loadfont.replace(/\'|\"/g,'').split(",");
		for (var i=0; i<fonts.length; i++) {
			checkLoad(fonts[i]);
		}
		
	} else {
		setTimeout( function(){ checkStyles(); } , 100);
	}
}

// 2. Check that all kernit fonts are loaded.
function checkLoad(font) {
	console.log(font+' checking');
	if (!loadTest(font)) {
		setTimeout( function(){ checkLoad(font) } , 100);
	} else {
		console.log(font+' loaded');
		fontRegister++;
		if (fontRegister >= fonts.length) {
			// 3. All fonts loaded, kern.
			console.log('Fonts loaded. Kerning... ');
			loadMessage('Fonts loaded. Kerning... ');
			setTimeout( function(){ kernIt(kernitClass); } , 100);// needs small delay to fire info message
		}
	}
}

// CHECK IF FONT IS LOADED AND READY TO RENDER
function loadTest(font) {
	var canv = document.createElement("canvas");
	document.body.appendChild(canv);
	canv.width = 100;
	canv.height = 100;
	
	var fc = canv.getContext("2d");
	fc.fillStyle = "red";
	fc.textBaseline = "top";
	
	// default width
	fc.font = defaultfont;
	var defaultWidth = fc.measureText('WMWMWMWMWMWMWMl').width;
	
	// loaded font width
	fc.font = '100px '+font+', Times, serif';
	fc.fillText('WMWMWMWMWMWMWMl', 0, 0);
	var letterWidth = fc.measureText('WMWMWMWMWMWMWMl').width;
	
	if (letterWidth != defaultWidth) {
		var imageData = fc.getImageData(0, 0, canv.width, canv.height);
		for (w=0; w<imageData.width; w++) {
			for (var h=0; h<imageData.height; h++) {
				if (imageData.data[((w*(imageData.width*4)) + (h*4))] > 0) {
					loadMessage(font+' loaded');
					var canvasTest = true;
					h=imageData.height;
					w=imageData.width;
				}
			}
		}
	}

	document.body.removeChild(canv);
	imageData = null;
	fc = null;
	canv = null;
	
	return ((defaultWidth != letterWidth) && canvasTest) ? true : false;
}

//////////////////////////////////////////////////////////////////////////////////////////

// IDENTIFY ALL TEXT TO BE KERNED
function kernIt(kc) {
	fontMeasure = $('body').append('<div id="fontMeasure">Font measure</div>');
	$('#fontMeasure').css({'position':'absolute', 'top':'-2000px', 'left':'0px'});
	
	kernitClass = (!kc) ? 'kernit' : kc;
	
	kernitTracking = parseInt(getStyle('.kernit_tracking', 'margin-right'));
	
	if (!kernitTracking && kernitTracking!=0) kernitTracking = 10;
	
	$('.'+kernitClass).each(function(index) {
		var elem = this;
		var pad = 0 - parseInt($(elem).parent().css('padding-left')) - parseInt($(elem).parent().css('padding-right'));
		setTimeout( function(){ kernEach( elem, pad ); }, 1 );
	});
		
	$('#fontMeasure').remove();
	
	if (returnFunc) returnFunc();
}

// INDIVIDUAL ELEMENT TO BE KERNED
function kernThis(targ, track, pad) {
	console.log("KERN: "+targ);

	fontMeasure = $('body').append('<div id="fontMeasure">Font measure</div>');
	$('#fontMeasure').css({'position':'absolute', 'top':'-2000px', 'left':'0px'});
	
	kernitTracking = (track) ? track : 10;
	if (!pad) pad = 0;
	
	$(targ).each(function(index) {
		var elem = this;
		setTimeout( function(){ kernEach( elem, pad ); }, 1 );
	});
	
	$('#fontMeasure').remove();
}

// KERN EACH TEXT ELEMENT - ALLOW FOR SCREEN UPDATE
function kernEach(elem, widthPadding) {
	var trans = '';
	var font = '';
	var fontsize = '';
	if (widthPadding == undefined) widthPadding = 10;
	
	kernitTag = $(elem).get(0).tagName.toLowerCase();
	
	// USE JQUERY TO GET FONT CHARACTERISTICS
	font = $(elem).css('font-family');
	fontsize = $(elem).css('font-size');
	if ($(elem).css('text-transform')) trans = $(elem).css('text-transform');
	
	// APPLY TEXT-STRANSFORM
	if (trans) trans = trans.replace(' ','');
	var txt = (trans=='uppercase') ? $(elem).text().toUpperCase() : (trans=='lowercase') ? $(elem).text().toLowerCase() : $(elem).text() ;
	
	$('#fontMeasure').append('<span id="fm" style="float:left; font-family:'+font+'; font-size:'+fontsize+'; color:red;">W</span>');
	var letterWidth = $('#fm').width();
	var textHeight = $('#fm').text(txt).height();
	
	// The #fontMeasure div does not work for the title page in FireFox.
	// Firefox is a pain in the bum. Try canvas measure instead...
	// CANVAS MEASURE
	var canv = document.createElement("canvas");
	var fc = canv.getContext("2d");
	fc.fillStyle = "red";
	fc.font = fontsize+' '+font;
	letterWidth = Math.round(fc.measureText('W').width);
	fc = null;
	canv = null;
	
	// REMOVE LETTER-SPACING
	$(elem).css('letter-spacing','0');
	
	// KERN TEXT
	if (kernitTag == 'span') $(elem).replaceWith(kernText(txt, font, fontsize, $(elem).attr('class').replace(kernitClass, '').replace(' ',' .'), letterWidth, textHeight))
	else $(elem).html(kernText(txt, font, fontsize, $(elem).attr('class').replace(kernitClass, '').replace(' ',' .'), letterWidth, textHeight));
	
	if ($(elem).attr('id')) fitText('#'+$(elem).attr('id'), $(elem).parent().width(), false);
	
	$(elem).css('visibility','visible');
		
	$('#fm').remove();
}

//////////////////////////////////////////////////////////////////////////////////////////

// KERN TEXT
function kernText(txt,fontface,fontsize,cssClass,letterWidth,textHeight) {
	// NOTES: The collision detection works best on larger type.
	// Instead of using the size of the font, the size is set to 100 for the collision detection.
	// The margin is then scaled to the size of the font.
	
	// store spacing for letter pairs
	var letters = {};
	textHeight = 120;
		
	// mozilla kerning on T is a bugger!
	// using float:left fixes it but allows word-breaking wrap
	var mozFloat = ($.browser.mozilla) ? 'float:left;' : '';
	
	var htm = '<span style="'+mozFloat+' margin:0;">'+txt.substr(0,1)+'</span>';
	for (var i=1; i<txt.length; i++) {
		
		if (txt.substr(i-1,1)===' ' || txt.substr(i,1)===' ' || txt.substr(i-1,1)==='_' || txt.substr(i,1)==='_' || txt.substr(i-1,1)===':' || txt.substr(i,1)===':') {
			// for pairs including a space - no collision detection
			var l1 = (txt.substr(i-1,1) === ' ') ? "spc" : txt.substr(i-1,1);
			var l2 = (txt.substr(i,1) === ' ') ?  "spc" : txt.substr(i,1) ;
			letters['ref'+(kernitTag ? kernitTag : '')+''+l1+''+l2] = 0;
		} else {
			// if letter pair has not been set, set it now
			var l1 = (txt.substr(i-1,1) === ' ') ? "H" : txt.substr(i-1,1);
			var l2 = (txt.substr(i,1) === ' ') ?  "H" : txt.substr(i,1) ;
			var m = 0;
		
			if (!letters.hasOwnProperty('ref'+l1+''+l2)) {
					
				// LOOK FOR CUSTOM KERNING BASED ON TAG + SPECIFIC CLASSES
				var cust = parseInt(getStyle(kernitTag+' .'+cssClass+' .k'+l1+' .k'+l2, 'margin-right'));
				if (isNaN(cust)) {
					cust = parseInt(getStyle(kernitTag+' .'+cssClass+' .k'+l1, 'margin-right'));
					if (isNaN(cust)) {
						cust = parseInt(getStyle(kernitTag+' .'+cssClass, 'margin-right'));
						// LOOK FOR CUSTOM KERNING BASED ON SPECIFIC CLASSES
						if (isNaN(cust)) {
							cust = parseInt(getStyle('.'+cssClass+' .k'+l1+' .k'+l2, 'margin-right'));
							if (isNaN(cust)) {
								cust = parseInt(getStyle('.'+cssClass+' .k'+l1, 'margin-right'));
								if (isNaN(cust)) {
									cust = parseInt(getStyle('.'+cssClass, 'margin-right'));
									// LOOK FOR CUSTOM KERNING BASED ON TAG AND KERNIT CLASS & LETTERS
									if (isNaN(cust)) {
										cust = parseInt(getStyle(kernitTag+' .'+kernitClass+' .k'+l1+' .k'+l2, 'margin-right'));
										if (isNaN(cust)) {
											cust = parseInt(getStyle(kernitTag+' .'+kernitClass+' .k'+l1, 'margin-right'));
											if (isNaN(cust)) {
												// LOOK FOR CUSTOM KERNING BASED ON KERNIT CLASS
												cust = parseInt(getStyle('.'+kernitClass+' .k'+l1+' .k'+l2, 'margin-right'));
												if (isNaN(cust)) {
													cust = parseInt(getStyle('.'+kernitClass+' .k'+l1, 'margin-right'));
													if (isNaN(cust)) {
														// LOOK FOR CUSTOM KERNING BASED ON JUST TAG AND KERNIT CLASS
														cust = parseInt(getStyle(kernitTag+' .'+kernitClass, 'margin-right'));
														if (isNaN(cust)) cust = kernitTracking;
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
								
				
				track = (cust!=null ? cust : kernitTracking) / 100 * letterWidth;
				
				var marg = getMargin(l1,l2,fontface,textHeight);
				m = (marg + track * retinaDisplay) / 100;
								
				// save letter pair spacing for reuse - only relevant for long passages of text...
				letters['ref'+(kernitTag ? kernitTag : '')+''+l1+''+l2] = m;
			}
					
		}
		if (l2=='spc') htm += '<span style="'+mozFloat+' margin:0 0 0 0;" class="space"> </span>'//&nbsp;
		else htm += '<span style="'+mozFloat+' margin:0 0 0 '+letters['ref'+(kernitTag ? kernitTag : '')+''+l1+''+l2]+'em;">'+(l2)+'</span>';//txt.substr(i,1)
	}
	return htm;
}

//////////////////////////////////////////////////////////////////////////////////////////

// GET SPACING FOR LETTER PAIRS
function getMargin(l1,l2,fontface,textHeight) {
	var canv = document.createElement("canvas");
	canv.width = window.innerWidth;
	canv.height = textHeight;
	
	var c = canv.getContext("2d");
		
	c.font = "100px "+fontface;
	
	c.fillStyle = "red";
	c.textBaseline = "top";
	
	var w1 = Math.round(c.measureText(l1).width)+10;
	var w2 = Math.round(c.measureText(l2).width);

	c.fillText(l1, 0, 0);
	var sprite1 = c.getImageData(0, 0, w1, textHeight);
	
	c.fillText(l2, w1, 0);
	var sprite2 = c.getImageData(w1, 0, w2, textHeight);
	
	var x = w1;
	
	while (!isPixelCollision( sprite1, 0, 0, sprite2, x, 0, false ) && x>-1) {
		x--;
	}
	
	c.fillStyle = "black";
	c.fillText(l2, x, 0);
	
	return x-c.measureText(l1).width;
}

//////////////////////////////////////////////////////////////////////////////////////////

/**
 * collission function by Joseph Lenton - PlayMyCode.com
 * http://www.playmycode.com/blog/2011/08/javascript-per-pixel-html5-canvas-image-collision-detection/
 *
 * @param first An ImageData object from the first image we are colliding with.
 * @param x The x location of 'first'.
 * @param y The y location of 'first'.
 * @param other An ImageData object from the second image involved in the collision check.
 * @param x2 The x location of 'other'.
 * @param y2 The y location of 'other'.
 * @param isCentred True if the locations refer to the centre of 'first' and 'other', false to specify the top left corner.
 */
function isPixelCollision( first, x, y, other, x2, y2, isCentred, overRide )
{
    // we need to avoid using floats, as were doing array lookups
    x  = Math.round( x );
    y  = Math.round( y );
    x2 = Math.round( x2 );
    y2 = Math.round( y2 );

    var w  = first.width,
        h  = first.height,
        w2 = other.width,
        h2 = other.height ;
        
    // deal with the image being centred
    if ( isCentred ) {
        // fast rounding, but positive only
        x  -= ( w/2 + 0.5) << 0
        y  -= ( h/2 + 0.5) << 0
        x2 -= (w2/2 + 0.5) << 0
        y2 -= (h2/2 + 0.5) << 0
    }

    // find the top left and bottom right corners of overlapping area
    var xMin = Math.max( x, x2 ),
        yMin = Math.max( y, y2 ),
        xMax = Math.min( x+w, x2+w2 ),
        yMax = Math.min( y+h, y2+h2 );

    // Sanity collision check, we ensure that the top-left corner is both
    // above and to the left of the bottom-right corner.
    if ( xMin >= xMax || yMin >= yMax ) {
        return false;
    }

    var xDiff = xMax - xMin,
        yDiff = yMax - yMin;

    // get the pixels out from the images
    var pixels  = first.data,
        pixels2 = other.data;

    // if the area is really small,
    // then just perform a normal image collision check
    if ( (xDiff < 4 && yDiff < 4) || overRide ) {
        for ( var pixelX = xMin; pixelX < xMax; pixelX++ ) {
            for ( var pixelY = yMin; pixelY < yMax; pixelY++ ) {
                if (
                        ( pixels [ ((pixelX-x ) + (pixelY-y )*w )*4 + 3 ] !== 0 ) &&
                        ( pixels2[ ((pixelX-x2) + (pixelY-y2)*w2)*4 + 3 ] !== 0 )
                ) {
                    return true;
                }
            }
        }
    } else {
        /* What is this doing?
         * It is iterating over the overlapping area,
         * across the x then y the,
         * checking if the pixels are on top of this.
         *
         * What is special is that it increments by incX or incY,
         * allowing it to quickly jump across the image in large increments
         * rather then slowly going pixel by pixel.
         *
         * This makes it more likely to find a colliding pixel early.
         */

        // Work out the increments,
        // it's a third, but ensure we don't get a tiny
        // slither of an area for the last iteration (using fast ceil).
        var incX = xDiff / 3.0,
            incY = yDiff / 3.0;
        incX = (~~incX === incX) ? incX : (incX+1 | 0);
        incY = (~~incY === incY) ? incY : (incY+1 | 0);

        for ( var offsetY = 0; offsetY < incY; offsetY++ ) {
            for ( var offsetX = 0; offsetX < incX; offsetX++ ) {
                for ( var pixelY = yMin+offsetY; pixelY < yMax; pixelY += incY ) {
                    for ( var pixelX = xMin+offsetX; pixelX < xMax; pixelX += incX ) {
                        if (
                                ( pixels [ ((pixelX-x ) + (pixelY-y )*w )*4 + 3 ] !== 0 ) &&
                                ( pixels2[ ((pixelX-x2) + (pixelY-y2)*w2)*4 + 3 ] !== 0 )
                        ) {
                            return true;
                        }
                    }
                }
            }
        }
    }

    return false;
}


//////////////////////////////////////////////////////////////////////////////////////////

function getCSSRule(ruleName, deleteFlag) {               // Return requested style obejct
/*    ruleName=ruleName.toLowerCase();                       // Convert test string to lower case. */
   if (document.styleSheets) {                            // If browser can play with stylesheets
      for (var i=0; i<document.styleSheets.length; i++) { // For each stylesheet
         var styleSheet=document.styleSheets[i];          // Get the current Stylesheet
         var ii=0;                                        // Initialize subCounter.
         var cssRule=false;                               // Initialize cssRule. 
         do {                                             // For each rule in stylesheet
            if (styleSheet.cssRules) {                    // Browser uses cssRules?
               cssRule = styleSheet.cssRules[ii];         // Yes --Mozilla Style
            } else {                                      // Browser usses rules?
               cssRule = styleSheet.rules[ii];            // Yes IE style. 
            }                                             // End IE check.
            if (cssRule)  {                               // If we found a rule...
               if (cssRule.selectorText/* .toLowerCase() */==ruleName) { //  match ruleName?
                  if (deleteFlag=='delete') {             // Yes.  Are we deleteing?
                     if (styleSheet.cssRules) {           // Yes, deleting...
                        styleSheet.deleteRule(ii);        // Delete rule, Moz Style
                     } else {                             // Still deleting.
                        styleSheet.removeRule(ii);        // Delete rule IE style.
                     }                                    // End IE check.
                     return true;                         // return true, class deleted.
                  } else {                                // found and not deleting.
                     return cssRule;                      // return the style object.
                  }                                       // End delete Check
               }                                          // End found rule name
            }                                             // end found cssRule
            ii++;                                         // Increment sub-counter
         } while (cssRule)                                // end While loop
      }                                                   // end For loop
   }                                                      // end styleSheet ability check
   return false;                                          // we found NOTHING!
}                                                         // end getCSSRule 

function killCSSRule(ruleName) {                          // Delete a CSS rule   
   return getCSSRule(ruleName,'delete');                  // just call getCSSRule w/delete flag.
}                                                         // end killCSSRule

function addCSSRule(ruleName) {                           // Create a new css rule
   if (document.styleSheets) {                            // Can browser do styleSheets?
      if (!getCSSRule(ruleName)) {                        // if rule doesn't exist...
         if (document.styleSheets[0].addRule) {           // Browser is IE?
            document.styleSheets[0].addRule(ruleName, null,0);      // Yes, add IE style
         } else {                                         // Browser is IE?
            document.styleSheets[0].insertRule(ruleName+' { }', 0); // Yes, add Moz style.
         }                                                // End browser check
      }                                                   // End already exist check.
   }                                                      // End browser ability check.
   return getCSSRule(ruleName);                           // return rule we just created.
}

//

function getStyle(selector, attr) {
	var rule = getCSSRule(selector)['cssText'];
	
	var re = new RegExp("\\b("+attr+")(\\W)+[^;]+");
	
	return (re.exec(rule)) ? re.exec(rule)[0].split(':')[1] : null;
}

//////////////////////////////////////////////////////////////////////////////////////////

// INFO WINDOW UTIITY
function loadMessage(msg) {
	if (infoTimer) clearTimeout(infoTimer);
	$('#loadMessage').remove();
	$('body').append('<div id="loadMessage">'+msg+'</div>');
	$('#loadMessage').click(function() {
		$(this).remove();
	});
	infoTimer = setTimeout(function(){ $('#loadMessage').remove(); }, 1000);
}